import pickle
from pathlib import Path
from typing import List

from loguru import logger

from gpd.models.dependent import Dependent


class DependentCache:
    path: Path
    owner: str
    project: str
    max_page: int

    def __init__(self, owner: str, project: str, max_page: int, directory: Path = Path("./cache")):
        self.owner = owner
        self.project = project
        self.max_page = max_page

        self.path = directory / "{}-{}-{}.pickle".format(self.owner, self.project, self.max_page)

    def dependents_from_file(self) -> List[Dependent]:
        try:
            with open(self.path, 'rb') as f:
                return pickle.load(f)
        except:
            logger.exception("path={path}", path=self.path)

    def dependents_to_file(self, dependents: List[Dependent]) -> None:
        if dependents:
            with open(self.path, 'wb') as f:
                pickle.dump(dependents, f)




from typing import List

from bs4 import BeautifulSoup, Tag

from gpd.models.dependent import Dependent
from gpd.parsers.dependent_parser import DependentParser


class TablePageParser:
    soup: BeautifulSoup

    def __init__(self, html: str):
        self.soup = BeautifulSoup(html, "html.parser")

    def get_dependents_estimate(self) -> int:
        estimate = self.soup.find("a", class_="btn-link selected").text.strip().replace(",", "")
        return int(estimate.split("\n")[0])

    def get_dependents(self) -> List[Dependent]:
        dependents = []
        for row in self.get_rows():
            dependents.append(Dependent(**DependentParser(row).to_dict()))
        return dependents

    def get_rows(self) -> List[Tag]:
        return self.soup.findAll("div", {"class": "Box-row"})

    def get_next_page_url(self) -> str:
        for tag in self.soup.find("div", {"class": "paginate-container"}).findAll('a'):
            if tag.text == "Next":
                return tag["href"]
        return ""



from bs4 import Tag


class DependentParser:
    dependent: Tag

    def __init__(self, table_row: Tag):
        self.dependent = table_row

    def get_name(self) -> str:
        return self.dependent.find('a', {"data-hovercard-type": "repository"}).text

    def get_stars(self) -> int:
        return int(self.dependent.findAll('span', class_="color-fg-muted text-bold pl-3")[0].text.strip().replace(",", ""))

    def get_forks(self) -> int:
        return int(self.dependent.findAll('span', class_="color-fg-muted text-bold pl-3")[1].text.strip().replace(",", ""))

    def get_author(self) -> str:
        return self.dependent.find('a', {"data-repository-hovercards-enabled": ""}).text

    def get_url(self) -> str:
        return self.urlify_for_github(self.get_href())

    def urlify_for_github(self, href: str):
        if href.find("http") == 0:
            return href
        return "https://github.com{}".format(href)

    def get_href(self) -> str:
        return self.dependent.find('a', {"data-hovercard-type": "repository"})["href"]

    def to_dict(self):
        return {"name": self.get_name(), "stars": self.get_stars(), "forks": self.get_forks(), "author": self.get_author(), "url": self.get_url()}



from typing import List, Dict, Any

from tabulate import tabulate

from gpd.models.dependent import Dependent


class DependentsStats:
    dependents: List[Dependent]

    def __init__(self, dependents: List[Dependent]):
        self.dependents = dependents

    def get_count(self):
        return len(self.dependents)

    def get_top_stars(self, max: int = 20) -> str:
        self.dependents.sort(key=lambda x: x.stars, reverse=True)
        return self._prettify_dependents(self.dependents[:max])

    def get_top_forks(self, max: int = 20) -> str:
        self.dependents.sort(key=lambda x: x.forks, reverse=True)
        return self._prettify_dependents(self.dependents[:max])

    def _prettify_dependents(self, dependents: List[Dependent]) -> str:
        return tabulate(self._group_dependents_by_key(dependents), headers="keys")

    def _group_dependents_by_key(self, dependents: List[Dependent]) -> Dict[str, Any]:
        groups = {}
        for dependent in dependents:
            for key, value in dependent.dict().items():
                if groups.get(key):
                    groups[key].append(value)
                else:
                    groups[key] = [value]
        return groups



"""
    gpd.py
    ------------------

    Runs the project.

    :copyright: 2019 MislavJaksic
    :license: MIT License
"""
import argparse
import pickle
import socket
import sys
import time
from typing import Dict, Any, List

import requests
from loguru import logger
from tqdm import tqdm

from gpd.dependent_cache import DependentCache
from gpd.dependents_stats import DependentsStats
from gpd.models.dependent import Dependent
from gpd.parsers.table_page_parser import TablePageParser


def main() -> int:
    """main() will be run if you run this script directly"""
    args = get_cli_arguments()

    setup_loguru()

    cache = DependentCache(args["owner"], args["name"], args["max_page"])
    dependents = cache.dependents_from_file()
    if not dependents:
        dependents = get_dependents(args["owner"], args["name"], args["max_page"])
        cache.dependents_to_file(dependents)

    stats = DependentsStats(dependents)
    print(stats.get_top_stars())

    return 0


def get_dependents(owner: str, project: str, max_page: int) -> List[Dependent]:
    page_number: int = 0
    next_url: str = 'https://github.com/{}/{}/network/dependents'.format(owner, project)
    request = None
    dependents: List[Dependent] = []
    estimated_pages = get_estimate_of_dependent_pages(next_url)
    try:
        logger.info("start_url={start_url}, max_page={max_page}, estimated_pages={estimated_pages}", start_url=next_url, max_page=max_page, estimated_pages=estimated_pages)
        for page_number in tqdm(range(min(estimated_pages, max_page))):
            request = requests.get(next_url)
            parser = TablePageParser(request.text)

            dependents += parser.get_dependents()
            next_url = parser.get_next_page_url()

            time.sleep(1.5)
            if not next_url:
                break
    except:
        logger.exception("start_url={start_url}, next_url={next_url}, page_number={page_number}, request={request}", start_url='https://github.com/{}/{}/network/dependents'.format(owner, project), next_url=next_url, page_number=page_number, request=request)
        if request:
            with open("latest-error.html", 'wb') as f:
                f.write(request.content)
    return dependents

def get_estimate_of_dependent_pages(url: str):
    request = requests.get(url)
    parser = TablePageParser(request.text)
    dependents_estimate = parser.get_dependents_estimate()
    return int(dependents_estimate * 1.2 / 30)

def get_cli_arguments() -> Dict[str, Any]:
    parser = argparse.ArgumentParser(description="",
                                     epilog="")

    parser.add_argument("-o", "--owner", required=True, help="Project owner. For example 'github' in 'https://github.com/github/advisory-database'.")
    parser.add_argument("-n", "--name", required=True, help="Project name. For example 'advisory-database' in 'https://github.com/github/advisory-database'.")
    parser.add_argument("-m", "--max_page", type=int, nargs='?', const=sys.maxsize, default=sys.maxsize, help="How many pages of dependents do you want to parse before finishing. Default is sys.maxsize, infinity.")

    return vars(parser.parse_args())


def setup_loguru() -> None:
    host = socket.gethostname()
    ip = socket.gethostbyname(host)

    config = {
        "handlers": [
            {"sink": sys.stderr, "diagnose": False,
             "format": '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{extra[host]}</cyan>:<cyan>{extra[ip]}</cyan> | <cyan>{process}</cyan>:<cyan>{thread}</cyan> | <cyan>{module}</cyan>:<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | <level>{message}</level>'},
            {"sink": "file.log", "retention": "7 days", "serialize": True},
        ],
        "extra": {"host": host, "ip": ip},
    }
    logger.configure(**config)  # type: ignore


def run() -> None:
    """Entry point for the runnable script."""
    sys.exit(main())


if __name__ == "__main__":
    """main calls run()."""
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gpd', 'gpd.models', 'gpd.parsers']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2',
 'loguru>=0.6.0,<0.7.0',
 'pydantic>=1.9.1,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'tqdm>=4.64.0,<5.0.0']

entry_points = \
{'console_scripts': ['gpd = gpd.runner:run']}

setup_kwargs = {
    'name': 'gpd',
    'version': '1.0.1',
    'description': 'Get a list of popular projects that use a library.',
    'long_description': "## GPD: Get Popular Dependents\n\nWant to know which popular projects use a GitHub library? GPD helps you find how popular projects use your favourite library.  \n\n## Install\n\n```\n$: pip install gpd\n```\n\n## Usage\n\n```\n$: gpd --help\nusage: gpd [-h] -o OWNER -n NAME [-m [MAX_PAGE]]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -o OWNER, --owner OWNER\n                        Project owner. For example 'github' in 'https://github.com/github/advisory-database'.\n  -n NAME, --name NAME  Project name. For example 'advisory-database' in 'https://github.com/github/advisory-database'.\n  -m [MAX_PAGE], --max_page [MAX_PAGE]\n                        How many pages of dependents do you want to parse before finishing. Default is sys.maxsize, infinity.\n```\n\n```\n$: poetry run gpd -o  samuelcolvin -n pydantic\nname                             stars    forks  author          url\n-----------------------------  -------  -------  --------------  ---------------------------------------------------------\nansible                          53318    21956  ansible         https://github.com/ansible/ansible\nspaCy                            23504     3840  explosion       https://github.com/explosion/spaCy\nray                              20707     3597  ray-project     https://github.com/ray-project/ray\njina                             14697     1919  jina-ai         https://github.com/jina-ai/jina\nrasa                             14068     4004  RasaHQ          https://github.com/RasaHQ/rasa\naiohttp                          12471     1768  aio-libs        https://github.com/aio-libs/aiohttp\nOpenBBTerminal                   12086     1295  OpenBB-finance  https://github.com/OpenBB-finance/OpenBBTerminal\ndgl                               9702     2287  dmlc            https://github.com/dmlc/dgl\npandas-profiling                  9037     1319  ydataai         https://github.com/ydataai/pandas-profiling\nfull-stack-fastapi-postgresql     9027     1622  tiangolo        https://github.com/tiangolo/full-stack-fastapi-postgresql\nlabel-studio                      8943     1065  heartexlabs     https://github.com/heartexlabs/label-studio\nludwig                            8340      986  ludwig-ai       https://github.com/ludwig-ai/ludwig\nPySyft                            8140     1818  OpenMined       https://github.com/OpenMined/PySyft\nsqlmodel                          7466      304  tiangolo        https://github.com/tiangolo/sqlmodel\nvaex                              7094      548  vaexio          https://github.com/vaexio/vaex\nrobotframework                    7008     1977  robotframework  https://github.com/robotframework/robotframework\nairbyte                           6924     1359  airbytehq       https://github.com/airbytehq/airbyte\ndoccano                           6230     1315  doccano         https://github.com/doccano/doccano\nedx-platform                      6017     3277  openedx         https://github.com/openedx/edx-platform\nDeepPavlov                        5754     1034  deepmipt        https://github.com/deepmipt/DeepPavlov\n```\n\nNote: you can only search libraries published on GitHub. For example, you cannot find which GitHub projects use Python's [asyncio](https://docs.python.org/3/library/asyncio.html).  \n\n## Development\n\n```\n# Note: Install Python 3\n# Update pip and install virtualenv (dependency encapsulator) and black (linter; IDE needs to be restarted)\n\n# Note: install Poetry for Linux\n$: curl -sSL https://install.python-poetry.org | python3 -\n\n# Note: install Poetry for Windows\n$: (Invoke-WebRequest -Uri https://install.python-poetry.org -UseBasicParsing).Content | py -\n\n$: curl -sSL https://install.python-poetry.org | python3 - --uninstall\n```\n\n```\n$: poetry install  # install all dependencies\n```\n\n### dist\n\n```\n$: pip install dist/gpd-1.0.1-py3-none.any.whl\n\n$: gpd\n```\n\n### docs\n\n```\n$: poetry shell\n$: cd docs\n# Note: review source/conf.py and source/index.rst\n$: make html\n# Note: see docs in docs/build/apidocs/index.html\n```\n\n### gpd\n\n```\n$: poetry run gpd --help\n```\n\n### tests\n\n```\n$: poetry run pytest --durations=0\n```\n\n```\n$: poetry run pytest --cov=gpd --cov-report=html tests\n# Note: see coverage report in htmlcov/index.html\n# Note: exclude directories from coverage report through .coveragerc\n```\n\n### poetry.lock\n\nDependencies, Python version and the virtual environment are managed by `Poetry`.\n\n```\n$: poetry search Package-Name\n$: poetry add [-D] Package-Name[==Package-Version]\n```\n\n### pyproject.toml\n\nDefine project entry point and metadata.\n\n### Linters\n\n```\n$: poetry run black .\n```\n\n### MyPy\n\n```\n$: poetry run mypy ./gpd ./tests\n```\n\n### cProfile\n\n```\n$: poetry run python ./gpd/profiler.py\n```\n\n### Build and publish\n\n```\n$: poetry build\n\n# Note: get the token from your PiPy account\n$: poetry config pypi-token.pypi PyPI-Api-Access-Token\n```\n\n```\n$: poetry publish --build\n```\n\n```\nhttps://pypi.org/project/gpd/\n```\n",
    'author': 'Mislav Jaksic',
    'author_email': 'jaksicmislav@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MislavJaksic/get-popular-dependents',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
